@@ -18,6 +18,7 @@ from .caching import (  # noqa: F401
 from .compression import compr
 from .registry import filesystem, get_filesystem_class
 from .utils import (
+    IOWrapper,
     _unstrip_protocol,
     build_name_function,
     infer_compression,
@@ -147,7 +148,11 @@ class OpenFile(object):
             closer()  # original close bound method of the final file-like
             _close(fobjects, mode)  # call close on other dependent file-likes
 
-        out.close = close
+        try:
+            out.close = close
+        except AttributeError:
+            out = IOWrapper(out, lambda: _close(fobjects, mode))
+
         return out
 
     def close(self):
@@ -528,7 +533,8 @@ def strip_protocol(urlpath):
 
 
 def expand_paths_if_needed(paths, mode, num, fs, name_function):
-    """Expand paths if they have a ``*`` in them.
+    """Expand paths if they have a ``*`` in them (write mode) or any of ``*?[]``
+    in them (read mode).
 
     :param paths: list of paths
     mode: str
@@ -544,23 +550,32 @@ def expand_paths_if_needed(paths, mode, num, fs, name_function):
     """
     expanded_paths = []
     paths = list(paths)
-    if "w" in mode and sum([1 for p in paths if "*" in p]) > 1:
-        raise ValueError("When writing data, only one filename mask can be specified.")
-    elif "w" in mode:
+
+    if "w" in mode:  # read mode
+        if sum([1 for p in paths if "*" in p]) > 1:
+            raise ValueError(
+                "When writing data, only one filename mask can be specified."
+            )
         num = max(num, len(paths))
-    for curr_path in paths:
-        if "*" in curr_path:
-            if "w" in mode:
+
+        for curr_path in paths:
+            if "*" in curr_path:
                 # expand using name_function
                 expanded_paths.extend(_expand_paths(curr_path, name_function, num))
             else:
+                expanded_paths.append(curr_path)
+        # if we generated more paths that asked for, trim the list
+        if len(expanded_paths) > num:
+            expanded_paths = expanded_paths[:num]
+
+    else:  # read mode
+        for curr_path in paths:
+            if has_magic(curr_path):
                 # expand using glob
                 expanded_paths.extend(fs.glob(curr_path))
-        else:
-            expanded_paths.append(curr_path)
-    # if we generated more paths that asked for, trim the list
-    if "w" in mode and len(expanded_paths) > num:
-        expanded_paths = expanded_paths[:num]
+            else:
+                expanded_paths.append(curr_path)
+
     return expanded_paths
 
 
