@@ -1,5 +1,4 @@
 import importlib
-from distutils.version import LooseVersion
 
 __all__ = ["registry", "get_filesystem_class", "default"]
 
@@ -104,6 +103,7 @@ known_implementations = {
         "err": 'HTTPFileSystem requires "requests" and "aiohttp" to be installed',
     },
     "zip": {"class": "fsspec.implementations.zip.ZipFileSystem"},
+    "tar": {"class": "fsspec.implementations.tar.TarFileSystem"},
     "gcs": {
         "class": "gcsfs.GCSFileSystem",
         "err": "Please install gcsfs to access Google Storage",
@@ -129,11 +129,21 @@ known_implementations = {
         "class": "fsspec.implementations.hdfs.PyArrowHDFS",
         "err": "pyarrow and local java libraries required for HDFS",
     },
+    "arrow_hdfs": {
+        "class": "fsspec.implementations.arrow.HadoopFileSystem",
+        "err": "pyarrow and local java libraries required for HDFS",
+    },
     "webhdfs": {
         "class": "fsspec.implementations.webhdfs.WebHDFS",
         "err": 'webHDFS access requires "requests" to be installed',
     },
     "s3": {"class": "s3fs.S3FileSystem", "err": "Install s3fs to access S3"},
+    "s3a": {"class": "s3fs.S3FileSystem", "err": "Install s3fs to access S3"},
+    "wandb": {"class": "wandbfs.WandbFS", "err": "Install wandbfs to access wandb"},
+    "oci": {
+        "class": "ocifs.OCIFileSystem",
+        "err": "Install ocifs to access OCI Object Storage",
+    },
     "adl": {
         "class": "adlfs.AzureDatalakeFileSystem",
         "err": "Install adlfs to access Azure Datalake Gen1",
@@ -183,10 +193,17 @@ known_implementations = {
         "err": "LibArchive requires to be installed",
     },
     "reference": {"class": "fsspec.implementations.reference.ReferenceFileSystem"},
+    "generic": {"class": "fsspec.generic.GenericFileSystem"},
+    "oss": {
+        "class": "ossfs.OSSFileSystem",
+        "err": "Install ossfs to access Alibaba Object Storage System",
+    },
+    "webdav": {
+        "class": "webdav4.fsspec.WebdavFileSystem",
+        "err": "Install webdav4 to access WebDAV",
+    },
 }
 
-minversions = {"s3fs": LooseVersion("0.3.0"), "gcsfs": LooseVersion("0.3.0")}
-
 
 def get_filesystem_class(protocol):
     """Fetch named protocol implementation from the registry
@@ -219,19 +236,20 @@ def get_filesystem_class(protocol):
 
 
 def _import_class(cls, minv=None):
-    mod, name = cls.rsplit(".", 1)
-    minv = minv or minversions
-    minversion = minv.get(mod, None)
+    """Take a string FQP and return the imported class or identifier
 
-    mod = importlib.import_module(mod)
-    if minversion:
-        version = getattr(mod, "__version__", None)
-        if version and LooseVersion(version) < minversion:
-            raise RuntimeError(
-                "'{}={}' is installed, but version '{}' or "
-                "higher is required".format(mod.__name__, version, minversion)
-            )
-    return getattr(mod, name)
+    clas is of the form "package.module.klass" or "package.module:subobject.klass"
+    """
+    if ":" in cls:
+        mod, name = cls.rsplit(":", 1)
+        mod = importlib.import_module(mod)
+        for part in name.split("."):
+            mod = getattr(mod, part)
+        return mod
+    else:
+        mod, name = cls.rsplit(".", 1)
+        mod = importlib.import_module(mod)
+        return getattr(mod, name)
 
 
 def filesystem(protocol, **storage_options):
@@ -242,3 +260,11 @@ def filesystem(protocol, **storage_options):
     """
     cls = get_filesystem_class(protocol)
     return cls(**storage_options)
+
+
+def available_protocols():
+    """Return a list of the implemented protocols.
+
+    Note that any given protocol may require extra packages to be importable.
+    """
+    return list(known_implementations)
