@@ -4,7 +4,6 @@ import os
 import threading
 import warnings
 import weakref
-from distutils.version import LooseVersion
 from errno import ESPIPE
 from glob import has_magic
 from hashlib import sha256
@@ -15,7 +14,7 @@ from .dircache import DirCache
 from .transaction import Transaction
 from .utils import (
     _unstrip_protocol,
-    get_package_version_without_import,
+    isfilelike,
     other_paths,
     read_block,
     stringify_path,
@@ -71,6 +70,7 @@ class _Cached(type):
             cls._cache.clear()
             cls._pid = os.getpid()
         if not skip and cls.cachable and token in cls._cache:
+            cls._latest = token
             return cls._cache[token]
         else:
             obj = super().__call__(*args, **kwargs)
@@ -84,24 +84,12 @@ class _Cached(type):
                 mirror_sync_methods(obj)
 
             if cls.cachable and not skip:
+                cls._latest = token
                 cls._cache[token] = obj
             return obj
 
 
-pa_version = get_package_version_without_import("pyarrow")
-if pa_version and LooseVersion(pa_version) < LooseVersion("2.0"):
-    try:
-        import pyarrow as pa
-
-        up = pa.filesystem.DaskFileSystem
-    except ImportError:  # pragma: no cover
-        # pyarrow exists but doesn't import for some reason
-        up = object
-else:  # pragma: no cover
-    up = object
-
-
-class AbstractFileSystem(up, metaclass=_Cached):
+class AbstractFileSystem(metaclass=_Cached):
     """
     An abstract super-class for pythonic file-systems
 
@@ -111,9 +99,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     cachable = True  # this class can be cached, instances reused
     _cached = False
-    blocksize = 2 ** 22
+    blocksize = 2**22
     sep = "/"
     protocol = "abstract"
+    _latest = None
     async_impl = False
     root_marker = ""  # For some FSs, may require leading '/' or other character
 
@@ -196,6 +185,17 @@ class AbstractFileSystem(up, metaclass=_Cached):
         # use of root_marker to make minimum required path, e.g., "/"
         return path or cls.root_marker
 
+    def unstrip_protocol(self, name):
+        """Format FS-specific path to generic, including protocol"""
+        if isinstance(self.protocol, str):
+            if name.startswith(self.protocol):
+                return name
+            return self.protocol + "://" + name
+        else:
+            if name.startswith(tuple(self.protocol)):
+                return name
+            return self.protocol[0] + "://" + name
+
     @staticmethod
     def _get_kwargs_from_urls(path):
         """If kwargs can be encoded in the paths, extract them here
@@ -211,14 +211,13 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     @classmethod
     def current(cls):
-        """Return the most recently created FileSystem
+        """Return the most recently instantiated FileSystem
 
         If no instance has been created, then create one with defaults
         """
-        if not len(cls._cache):
-            return cls()
-        else:
-            return list(cls._cache.values())[-1]
+        if cls._latest in cls._cache:
+            return cls._cache[cls._latest]
+        return cls()
 
     @property
     def transaction(self):
@@ -293,7 +292,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         path: str
             leaf directory name
         exist_ok: bool (False)
-            If True, will error if the target already exists
+            If False, will error if the target already exists
         """
         pass  # not necessary to implement, may not have directories
 
@@ -424,7 +423,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         for d in full_dirs:
             yield from self.walk(d, maxdepth=maxdepth, detail=detail, **kwargs)
 
-    def find(self, path, maxdepth=None, withdirs=False, **kwargs):
+    def find(self, path, maxdepth=None, withdirs=False, detail=False, **kwargs):
         """List all files below path.
 
         Like posix ``find`` command without conditions
@@ -442,7 +441,6 @@ class AbstractFileSystem(up, metaclass=_Cached):
         # TODO: allow equivalent of -name parameter
         path = self._strip_protocol(path)
         out = dict()
-        detail = kwargs.pop("detail", False)
         for _, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
             if withdirs:
                 files.update(dirs)
@@ -636,6 +634,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Size in bytes of file"""
         return self.info(path).get("size", None)
 
+    def sizes(self, paths):
+        """Size in bytes of each file in a list of paths"""
+        return [self.size(p) for p in paths]
+
     def isdir(self, path):
         """Is this entry directory-like?"""
         try:
@@ -755,20 +757,29 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             return self.cat_file(paths[0], **kwargs)
 
-    def get_file(self, rpath, lpath, callback=_DEFAULT_CALLBACK, **kwargs):
+    def get_file(
+        self, rpath, lpath, callback=_DEFAULT_CALLBACK, outfile=None, **kwargs
+    ):
         """Copy single remote file to local"""
-        if self.isdir(rpath):
-            os.makedirs(lpath, exist_ok=True)
-            return None
+        if isfilelike(lpath):
+            outfile = lpath
+        else:
+            if self.isdir(rpath):
+                os.makedirs(lpath, exist_ok=True)
+                return None
+
+            if outfile is None:
+                outfile = open(lpath, "wb")
 
         with self.open(rpath, "rb", **kwargs) as f1:
             callback.set_size(getattr(f1, "size", None))
-            with open(lpath, "wb") as f2:
-                data = True
-                while data:
-                    data = f1.read(self.blocksize)
-                    segment_len = f2.write(data)
-                    callback.relative_update(segment_len)
+            data = True
+            while data:
+                data = f1.read(self.blocksize)
+                segment_len = outfile.write(data)
+                callback.relative_update(segment_len)
+        if not isfilelike(lpath):
+            outfile.close()
 
     def get(self, rpath, lpath, recursive=False, callback=_DEFAULT_CALLBACK, **kwargs):
         """Copy file(s) to local.
@@ -799,13 +810,13 @@ class AbstractFileSystem(up, metaclass=_Cached):
             return None
 
         with open(lpath, "rb") as f1:
-            callback.set_size(f1.seek(0, 2))
+            size = f1.seek(0, 2)
+            callback.set_size(size)
             f1.seek(0)
 
             self.mkdirs(self._parent(os.fspath(rpath)), exist_ok=True)
             with self.open(rpath, "wb", **kwargs) as f2:
-                data = True
-                while data:
+                while f1.tell() < size:
                     data = f1.read(self.blocksize)
                     segment_len = f2.write(data)
                     callback.relative_update(segment_len)
@@ -840,12 +851,12 @@ class AbstractFileSystem(up, metaclass=_Cached):
             self.put_file(lpath, rpath, **kwargs)
 
     def head(self, path, size=1024):
-        """ Get the first ``size`` bytes from file """
+        """Get the first ``size`` bytes from file"""
         with self.open(path, "rb") as f:
             return f.read(size)
 
     def tail(self, path, size=1024):
-        """ Get the last ``size`` bytes from file """
+        """Get the last ``size`` bytes from file"""
         with self.open(path, "rb") as f:
             f.seek(max(-size, -f.size), 2)
             return f.read()
@@ -907,7 +918,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         return list(sorted(out))
 
     def mv(self, path1, path2, recursive=False, maxdepth=None, **kwargs):
-        """ Move file(s) from one location to another """
+        """Move file(s) from one location to another"""
         self.copy(path1, path2, recursive=recursive, maxdepth=maxdepth)
         self.rm(path1, recursive=recursive)
 
@@ -968,7 +979,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
             **kwargs,
         )
 
-    def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
+    def open(
+        self,
+        path,
+        mode="rb",
+        block_size=None,
+        cache_options=None,
+        compression=None,
+        **kwargs,
+    ):
         """
         Return a file-like object from the filesystem
 
@@ -985,6 +1004,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
             Some indication of buffering - this is a value in bytes
         cache_options : dict, optional
             Extra arguments to pass through to the cache.
+        compression: string or None
+            If given, open file using compression codec. Can either be a compression
+            name (a key in ``fsspec.compression.compr``) or "infer" to guess the
+            compression from the filename suffix.
         encoding, errors, newline: passed on to TextIOWrapper for text mode
         """
         import io
@@ -999,7 +1022,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 if k in kwargs
             }
             return io.TextIOWrapper(
-                self.open(path, mode, block_size, **kwargs), **text_kwargs
+                self.open(
+                    path,
+                    mode,
+                    block_size=block_size,
+                    cache_options=cache_options,
+                    compression=compression,
+                    **kwargs,
+                ),
+                **text_kwargs,
             )
         else:
             ac = kwargs.pop("autocommit", not self._intrans)
@@ -1011,6 +1042,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 cache_options=cache_options,
                 **kwargs,
             )
+            if compression is not None:
+                from fsspec.compression import compr
+                from fsspec.core import get_compression
+
+                compression = get_compression(path, compression)
+                compress = compr[compression]
+                f = compress(f, mode=mode[0])
+
             if not ac and "r" not in mode:
                 self.transaction.files.append(f)
             return f
@@ -1142,7 +1181,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         # all instances already also derive from pyarrow
         return self
 
-    def get_mapper(self, root, check=False, create=False):
+    def get_mapper(self, root="", check=False, create=False, missing_exceptions=None):
         """Create key/value store based on this file-system
 
         Makes a MutableMapping interface to the FS at the given root path.
@@ -1150,7 +1189,13 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """
         from .mapping import FSMap
 
-        return FSMap(root, self, check, create)
+        return FSMap(
+            root,
+            self,
+            check=check,
+            create=create,
+            missing_exceptions=missing_exceptions,
+        )
 
     @classmethod
     def clear_instance_cache(cls):
@@ -1263,7 +1308,7 @@ class AbstractBufferedFile(io.IOBase):
     ``_initiate_upload`` and ``_fetch_range``.
     """
 
-    DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
+    DEFAULT_BLOCK_SIZE = 5 * 2**20
     _details = None
 
     def __init__(
@@ -1389,14 +1434,14 @@ class AbstractBufferedFile(io.IOBase):
         """Throw away temporary file"""
 
     def info(self):
-        """ File information about this path """
+        """File information about this path"""
         if "r" in self.mode:
             return self.details
         else:
             raise ValueError("Info not available while writing")
 
     def tell(self):
-        """ Current file location """
+        """Current file location"""
         return self.loc
 
     def seek(self, loc, whence=0):
@@ -1503,7 +1548,7 @@ class AbstractBufferedFile(io.IOBase):
         # may not yet have been initialized, may need to call _initialize_upload
 
     def _initiate_upload(self):
-        """ Create remote file/upload """
+        """Create remote file/upload"""
         pass
 
     def _fetch_range(self, start, end):
