@@ -7,9 +7,11 @@ import sys
 from contextlib import contextmanager
 from functools import partial
 from hashlib import md5
+from types import TracebackType
+from typing import IO, AnyStr, Callable, Iterable, Iterator, List, Optional, Type
 from urllib.parse import urlsplit
 
-DEFAULT_BLOCK_SIZE = 5 * 2 ** 20
+DEFAULT_BLOCK_SIZE = 5 * 2**20
 
 
 def infer_storage_options(urlpath, inherit_storage_options=None):
@@ -241,14 +243,14 @@ def read_block(f, offset, length, delimiter=None, split_before=False):
     """
     if delimiter:
         f.seek(offset)
-        found_start_delim = seek_delimiter(f, delimiter, 2 ** 16)
+        found_start_delim = seek_delimiter(f, delimiter, 2**16)
         if length is None:
             return f.read()
         start = f.tell()
         length -= start - offset
 
         f.seek(start + length)
-        found_end_delim = seek_delimiter(f, delimiter, 2 ** 16)
+        found_end_delim = seek_delimiter(f, delimiter, 2**16)
         end = f.tell()
 
         # Adjust split location to before delimiter iff seek found the
@@ -384,6 +386,13 @@ def is_exception(obj):
     return isinstance(obj, BaseException)
 
 
+def isfilelike(f):
+    for attr in ["read", "close", "tell"]:
+        if not hasattr(f, attr):
+            return False
+    return True
+
+
 def get_protocol(url):
     parts = re.split(r"(\:\:|\://)", url, 1)
     if len(parts) > 1:
@@ -455,14 +464,7 @@ def setup_logging(logger=None, logger_name=None, level="DEBUG", clear=True):
 
 
 def _unstrip_protocol(name, fs):
-    if isinstance(fs.protocol, str):
-        if name.startswith(fs.protocol):
-            return name
-        return fs.protocol + "://" + name
-    else:
-        if name.startswith(tuple(fs.protocol)):
-            return name
-        return fs.protocol[0] + "://" + name
+    return fs.unstrip_protocol(name)
 
 
 def mirror_from(origin_name, methods):
@@ -540,3 +542,80 @@ def merge_offset_ranges(paths, starts, ends, max_gap=0, max_block=None, sort=Tru
 
     # `paths` is empty. Just return input lists
     return paths, starts, ends
+
+
+class IOWrapper(IO):
+    """Wrapper for a file-like object that can be used in situations where we might
+    want to, e.g., monkey-patch the close method but can't.
+    (cf https://github.com/fsspec/filesystem_spec/issues/725)
+    """
+
+    def __init__(self, fp: IO, closer: Callable[[], None]):
+        self.fp = fp
+        self.closer = closer
+
+    def close(self) -> None:
+        self.fp.close()
+
+    def fileno(self) -> int:
+        return self.fp.fileno()
+
+    def flush(self) -> None:
+        self.fp.flush()
+
+    def isatty(self) -> bool:
+        return self.fp.isatty()
+
+    def read(self, n: int = ...) -> AnyStr:
+        return self.fp.read(n)
+
+    def readable(self) -> bool:
+        return self.fp.readable()
+
+    def readline(self, limit: int = ...) -> AnyStr:
+        return self.fp.readline(limit)
+
+    def readlines(self, hint: int = ...) -> List[AnyStr]:
+        return self.fp.readlines(hint)
+
+    def seek(self, offset: int, whence: int = ...) -> int:
+        return self.fp.seek(offset, whence)
+
+    def seekable(self) -> bool:
+        return self.fp.seekable()
+
+    def tell(self) -> int:
+        return self.fp.tell()
+
+    def truncate(self, size: Optional[int] = ...) -> int:
+        return self.fp.truncate(size)
+
+    def writable(self) -> bool:
+        return self.fp.writable()
+
+    def write(self, s: AnyStr) -> int:
+        return self.fp.write(s)
+
+    def writelines(self, lines: Iterable[AnyStr]) -> None:
+        self.fp.writelines(lines)
+
+    def __next__(self) -> AnyStr:
+        return next(self.fp)
+
+    def __iter__(self) -> Iterator[AnyStr]:
+        return iter(self.fp)
+
+    def __enter__(self) -> IO[AnyStr]:
+        return self.fp.__enter__()
+
+    def __exit__(
+        self,
+        t: Optional[Type[BaseException]],
+        value: Optional[BaseException],
+        traceback: Optional[TracebackType],
+    ) -> Optional[bool]:
+        return self.fp.__exit__(t, value, traceback)
+
+    # forward anything else too
+    def __getattr__(self, name):
+        return getattr(self.fp, name)
